# coding: utf-8
from __future__ import absolute_import

import argparse
import importlib
import logging
import os
import pickle
import pkg_resources
import re
import shelve
import six
import threading
import time
import yaml

from . import utils

__version__ = pkg_resources.get_distribution(__name__).version


# pass this object to callback, to terminate the bot
EXIT = object()


class Request(object):
    def __init__(self, message):
        self.message = message

    def respond(self, message):
        raise NotImplementedError('You have to implement \'respond\' method in you Request class.')


class Adapter(object):
    def __init__(self, bot, callback):
        self.bot = bot
        self.callback = callback

    def start(self):
        """You have to override this method, if you plugin requires some background activity.

        Here you can create a thread, but don't forget to make its `daemon` attrubute equal to True.
        """


class Plugin(object):
    def __init__(self, bot):
        self.bot = bot
        self.storage = self.bot.storage.with_prefix(self.__module__ + ':')

    def get_callbacks(self):
        for name in dir(self):
            value = getattr(self, name)
            if callable(value):
                patterns = getattr(value, '_patterns', [])
                for pattern in patterns:
                    yield (pattern, value)


class ThreadedPlugin(Plugin):
    """ThreadedPlugin allows you to do some processing in a background thread.

    This class will take care on proper thread execution and termination.

    * First, implement method `do_job`, which will be executed with given interval.
    * Then run method `self.start_worker(interval=60)` to run a thread.
      It will call your do_job callback each 60 seconds.
    * To stop job execution, call `self.stop_worker()`.

    See `thebot-instagram`, as an example.
    """
    def do_job(self):
        raise NotImplemented('Implement "do_job" method to get real work done.')

    def is_working(self):
        thread = getattr(self, '_thread', None)
        return thread is not None and thread.is_alive()

    def start_worker(self, interval=60):
        if self.is_working():
            return

        self._event = threading.Event()
        self._thread = threading.Thread(target=self._worker, kwargs=dict(interval=interval))
        self._thread.daemon = True
        self._thread.start()

    def stop_worker(self, wait=True):
        event = getattr(self, '_event', None)
        if event is not None:
            event.set()

        if wait:
            self._thread.join()

    def _worker(self, interval=60):
        countdown = 0
        logger = logging.getLogger('thebot.' + self.__class__.__name__)

        on_start = getattr(self, 'on_start', None)
        if on_start is not None:
            on_start()

        while not self._event.is_set():
            if countdown == 0:
                try:
                    self.do_job()
                except Exception:
                    logger.exception('Error during the task execution')

                countdown = interval
            else:
                countdown -= 1

            time.sleep(1)

        on_stop = getattr(self, 'on_stop', None)
        if on_stop is not None:
            on_stop()


def route(pattern):
    """Decorator to assign routes to plugin's methods.
    """
    def deco(func):
        if getattr(func, '_patterns', None) is None:
            func._patterns = []
        func._patterns.append(pattern)
        return func
    return deco


class HelpPlugin(Plugin):
    @route('help')
    def help(self, request):
        """Shows a help."""
        lines = []
        for pattern, callback in self.bot.patterns:
            docstring = callback.__doc__
            if docstring:
                lines.append(six.u('  {} — {}').format(pattern, docstring))
            else:
                lines.append(six.u('  ') + pattern)

        lines.sort()
        lines.insert(0, six.u('I support following commands:'))

        request.respond(six.u('\n').join(lines))


class Pickler(pickle.Pickler):
    def __init__(self, file, protocol=None, global_objects=None):
        pickle.Pickler.__init__(self, file, protocol=protocol)
        self._global_objects = dict(
            (id(obj), persistent_id)
            for persistent_id, obj in (global_objects or {}).items()
        )

    def persistent_id(self, obj):
        return self._global_objects.get(id(obj))


class Unpickler(pickle.Unpickler):
    def __init__(self, file, global_objects=None):
        pickle.Unpickler.__init__(self, file)
        self._global_objects = global_objects or {}

    def persistent_load(self, obj_id):
        return self._global_objects[obj_id]


class Shelve(shelve.DbfilenameShelf):
    """A custom Shelve, to use custom Pickler and Unpickler.
    """
    def __init__(self, filename, global_objects=None):
        shelve.DbfilenameShelf.__init__(self, filename)
        self._global_objects = global_objects or {}

    def __getitem__(self, key):
        try:
            value = self.cache[key]
        except KeyError:
            f = six.BytesIO(self.dict[key])
            value = Unpickler(f, global_objects=self._global_objects).load()
            if self.writeback:
                self.cache[key] = value
        return value

    def __setitem__(self, key, value):
        if self.writeback:
            self.cache[key] = value
        f = six.BytesIO()
        p = Pickler(f, self._protocol, global_objects=self._global_objects)
        p.dump(value)
        self.dict[key] = f.getvalue()


class Storage(object):
    def __init__(self, filename, prefix='', global_objects=None):
        """Specials are used to restore references to some nonserializable objects,
        such as TheBot itself.
        """
        if isinstance(filename, Shelve):
            self._shelve = filename
        else:
            self._shelve = Shelve(filename, global_objects=global_objects)

        self.prefix = prefix
        self.global_objects = global_objects or {}

    def __getitem__(self, name):
        return self._shelve.__getitem__(self.prefix + name)

    def get(self, name, default=None):
        return self._shelve.get(self.prefix + name, default)

    def __setitem__(self, name, value):
        return self._shelve.__setitem__(self.prefix + name, value)

    def __delitem__(self, name):
        return self._shelve.__delitem__(self.prefix + name)

    def keys(self):
        return list(filter(lambda x: x.startswith(self.prefix), self._shelve.keys()))

    def clear(self):
        for key in self.keys():
            del self._shelve[key]

    def with_prefix(self, prefix):
        return Storage(self._shelve, prefix=prefix, global_objects=self.global_objects)

    def close(self):
        self._shelve.close()


class Config(object):
    def __init__(self):
        self._data = {}

    def read_from_file(self, filename):
        with open(filename) as f:
            self.read_from_string(f.read())

    def read_from_string(self, data):
        data = yaml.load(data)

        # This complex code allows you
        # to translate such YAML config:
        #
        # adapters: [xmpp, http]
        # xmpp:
        #   jid: thebot@ya.ru
        # http:
        #   port: 9000
        #
        # into a config with attributes `xmpp_jid` and `http_port`.
        for key in ('plugins', 'adapters'):
            for item in data.get(key, []):
                item_settings = data.pop(item, None)
                if item_settings is not None:
                    for k, v in item_settings.items():
                        data[item + '_' + k] = v


        self._data.update(data)

    def update_from_dict(self, args):
        """Gets settings from the Namespace object, generated by argparse."""
        for key, value in args.items():
            if value is not None:
                if key in ('adapters', 'plugins'):
                    value = value.split(',')

                self._data[key] = value

    def __repr__(self):
        return 'Config: {}'.format(self._data)

    def __getattr__(self, name):
        return self._data[name]


class Bot(object):
    def __init__(self, command_line_args=[], adapters=None, plugins=None, config_dict={}):
        self.adapters = []
        self.plugins = []
        self.patterns = []
        self.exiting = False

        def load(value, cls='Adapter'):
            """Returns class by it's name.

            Given a 'irc' string it will try to load the following:

            1) from thebot_irc import Adapter
            2) from thebot.batteries.irc import Adapter

            If all of them fail, it will raise ImportError
            """
            if isinstance(value, six.string_types):
                try:
                    module = importlib.import_module('thebot_' + value)
                except ImportError:
                    module = importlib.import_module('thebot.batteries.' + value)

                loaded_value = getattr(module, cls)
                if not hasattr(loaded_value, 'name'):
                    loaded_value.name = value
                return loaded_value

            return value


        config_filename = 'thebot.conf'

        default_config = """
            unittest: no
        """

        # read config first time, to figure out possible adapter
        config = Config()
        config.read_from_string(default_config)

        parser = Bot.get_general_options()
        defaults = dict(parser.parse_args([])._get_kwargs())
        config.update_from_dict(defaults)

        if os.path.exists(config_filename):
            config.read_from_file(config_filename)

        args, unknown = parser.parse_known_args(
            list(filter(lambda x: x not in ('--help', '-h'), command_line_args))
        )
        # we override options only if there was specified value different from default
        args = dict(
            (key, value)
            for key, value in args._get_kwargs()
                if value != defaults.get(key)
        )
        config.update_from_dict(args)


        # now, load plugin and adapter classes, collect their options
        # and parse command line again

        if adapters is None:
            adapter_classes = map(lambda a: load(a, 'Adapter'), config.adapters)
        else:
            # we've got adapters argument (it is used for testing purpose
            adapter_classes = adapters

        if plugins is None:
            plugin_classes = [load(a, 'Plugin') for a in  config.plugins]
        else:
            # we've got adapters argument (it is used for testing purpose
            plugin_classes = plugins

        plugin_classes.append(HelpPlugin)

        for cls in adapter_classes + plugin_classes:
            if hasattr(cls, 'get_options'):
                cls.get_options(parser)


        # now, reread config to write there all defaults from all plugins
        config = Config()
        config.read_from_string(default_config)

        # getting options' defaults from parser
        defaults = dict(parser.parse_args([])._get_kwargs())
        config.update_from_dict(defaults)

        # now applying settings from the config file
        if os.path.exists(config_filename):
            config.read_from_file(config_filename)

        # and finally, override them with command line options
        # if there was specified value different from default
        args = parser.parse_args(command_line_args)
        args = dict(
            (key, value)
            for key, value in args._get_kwargs()
                if value != defaults.get(key)
        )
        config.update_from_dict(args)
        self.config = config

        for key, value in config_dict.items():
            setattr(self.config, key, value)

        self.storage = Storage(self.config.storage_filename, global_objects=dict(bot=self))

        logging.basicConfig(
            filename=self.config.log_filename,
            format='[%(asctime)s] %(levelname)s %(name)s: %(message)s',
            level=logging.DEBUG if self.config.verbose else logging.WARNING,
        )

        # adapters and plugins initialization

        for adapter in adapter_classes:
            a = adapter(self, callback=self.on_request)
            a.start()
            self.adapters.append(a)

        for plugin_cls in plugin_classes:
            p = plugin_cls(self)
            self.plugins.append(p)
            self.patterns.extend(p.get_callbacks())

        self.patterns = [
            (utils.force_unicode(pattern), callback)
            for pattern, callback in self.patterns
        ]


    @staticmethod
    def get_general_options():
        parser = argparse.ArgumentParser(
            description='The Bot — Hubot\'s killer.'
        )
        parser.add_argument(
            '--verbose', '-v', action='store_true', default=False,
            help='Show more output. Default: False.'
        )
        parser.add_argument(
            '--log-filename', default='thebot.log',
            help='Log\'s filename. Default: thebot.log.'
        )
        parser.add_argument(
            '--storage-filename', default='thebot.storage',
            help='Path to a database file, used for TheBot\'s memory. Default: thebot.storage.'
        )

        group = parser.add_argument_group('General options')
        group.add_argument(
            '--adapters', '-a', default='console',
            help='Adapters to use. You can specify a comma-separated list to use more than one adapter. Default: console.',
        )
        group.add_argument(
            '--plugins', '-p', default='image,math,todo',
            help='Plugins to use. You can specify a comma-separated list to use more than one plugin. Default: image.',
        )
        return parser

    def on_request(self, request):
        if request is EXIT:
            self.exiting = True
        else:
            for pattern, callback in self.patterns:
                match = re.match(pattern, request.message)
                if match is not None:
                    result = callback(request, **match.groupdict())
                    if result is not None:
                        raise RuntimeError('Plugin {0} should not return response directly. Use request.respond(some message).')
                    break
            else:
                request.respond('I don\'t know command "{0}".'.format(request.message))

    def close(self):
        """Will close all connections here.
        """
        self.storage.close()

    def get_adapter(self, name):
        """Returns adapter by it's name."""
        for adapter in self.adapters:
            if getattr(adapter, 'name', None) == name:
                return adapter
        raise KeyError(six.u('Adapter {} not found').format(name))

    def get_plugin(self, name):
        """Returns plugin by it's name."""
        for plugin in self.plugins:
            if getattr(plugin, 'name', None) == name:
                return plugin
        raise KeyError(six.u('Plugin {} not found').format(name))




# coding: utf-8

import times
import datetime
import mock
import thebot
import six
import sys

from thebot import Request, Adapter, Plugin, Storage, route, Config
from thebot.batteries import todo
from nose.tools import eq_, assert_raises
from contextlib import closing


class Bot(thebot.Bot):
    """Test bot which uses slightly different settings."""
    def __init__(self, *args, **kwargs):
        python_version = '-'.join(map(str, sys.version_info[:3]))
        kwargs['config_dict'] = dict(
            unittest=True,
            log_filename='unittest.log',
            storage_filename='unittest-{}.storage'.format(python_version),
        )
        super(Bot, self).__init__(*args, **kwargs)

        self.storage.clear()


class TestRequest(Request):
    def __init__(self, message, bot, user):
        super(TestRequest, self).__init__(message)
        self.bot = bot
        self.user = user

    def respond(self, message):
        adapter = self.bot.get_adapter('test')
        adapter._lines.append(message)

    def get_user(self):
        return self.user


class TestAdapter(Adapter):
    name = 'test'

    def __init__(self, *args, **kwargs):
        super(TestAdapter, self).__init__(*args, **kwargs)
        self._lines = []

    def write(self, input_line, user='some user'):
        """This method is for test purpose.
        """
        self.callback(TestRequest(input_line, self.bot, user))


class TestPlugin(Plugin):
    @route('show me a cat')
    def show_a_cat(self, request):
        """Shows a cat."""
        request.respond('the Cat')

    @route('find (?P<this>.*)')
    def find(self, request, this=None):
        """Making a fake search of the term."""
        request.respond('I found {0}'.format(this))


def test_install_adapters():
    with closing(Bot(adapters=[TestAdapter], plugins=[])) as bot:
        assert len(bot.adapters) == 1


def test_install_plugins():
    with closing(Bot(adapters=[], plugins=[TestPlugin])) as bot:
        eq_(0, len(bot.adapters))
        eq_(2, len(bot.plugins)) # Help plugin is added by default
        eq_(3, len(bot.patterns))


def test_one_line():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.get_adapter('test')

        eq_(adapter._lines, [])
        adapter.write('show me a cat')
        eq_(adapter._lines, ['the Cat'])

        adapter.write('find Umputun')
        eq_(adapter._lines[-1], 'I found Umputun')


def test_unknown_command():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.adapters[0]

        eq_(adapter._lines, [])
        adapter.write('some command')
        eq_(adapter._lines, ['I don\'t know command "some command".'])


def test_exception_raised_if_plugin_returns_not_none():
    class BadPlugin(Plugin):
        @route('^do$')
        def do(self, request):
            return 'Hello world'


    with closing(Bot(adapters=[TestAdapter], plugins=[BadPlugin])) as bot:
        adapter = bot.adapters[0]

        assert_raises(RuntimeError, adapter.write, 'do')


def test_simple_storage():
    storage = Storage('/tmp/thebot.storage')
    storage.clear()

    eq_([], storage.keys())

    storage['blah'] = 'minor'
    storage['one'] = {'some': 'dict'}

    eq_(['blah', 'one'], sorted(storage.keys()))
    eq_('minor', storage['blah'])


def test_storage_nesting():
    storage = Storage('/tmp/thebot.storage')
    storage.clear()

    first = storage.with_prefix('first:')
    second = storage.with_prefix('second:')

    eq_([], storage.keys())

    first['blah'] = 'minor'
    second['one'] = {'some': 'dict'}

    eq_(['first:blah', 'second:one'], sorted(storage.keys()))
    eq_(['first:blah'], sorted(first.keys()))
    eq_(['second:one'], sorted(second.keys()))

    eq_('minor', first['blah'])
    assert_raises(KeyError, lambda: second['blah'])

    first.clear()
    eq_(['second:one'], sorted(storage.keys()))


def test_help_command():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        adapter = bot.adapters[0]

        adapter.write('help')
        eq_(
            [
                six.u('I support following commands:\n'
                      '  find (?P<this>.*) — Making a fake search of the term.\n'
                      '  help — Shows a help.\n'
                      '  show me a cat — Shows a cat.'
                )
            ],
            adapter._lines
        )


def test_delete_from_storage():
    storage = Storage('/tmp/thebot.storage')
    storage.clear()

    storage['blah'] = 'minor'
    del storage['blah']

    eq_([], sorted(storage.keys()))



def test_storage_restores_bot_attribute():
    with closing(Bot(adapters=[TestAdapter], plugins=[TestPlugin])) as bot:
        storage = Storage('/tmp/thebot.storage', global_objects=dict(bot=bot))
        storage.clear()

        original = Request('blah')
        original.bot = bot

        storage['request'] = original

        restored = storage['request']
        eq_(restored.bot, original.bot)


def test_storage_with_prefix_keeps_global_objects():
    storage = Storage('/tmp/thebot.storage', global_objects=dict(some='value'))
    prefixed = storage.with_prefix('nested:')

    eq_(storage.global_objects, prefixed.global_objects)


def test_get_adapter_by_name():
    with closing(Bot(adapters=[TestAdapter])) as bot:
        adapter = bot.get_adapter('test')
        assert isinstance(adapter, TestAdapter)


def test_todo_plugin():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:
        adapter = bot.get_adapter('test')

        adapter.write('remind at 2012-10-05 to Celebrate my birthday')
        adapter.write('remind at 2012-12-18 to Celebrate daughter\'s birthday')
        adapter.write('remind at 2012-09-01 to Write a doc for TheBot')
        adapter.write('my tasks')

        eq_(
            [
                'ok',
                'ok',
                'ok',
                '16) 2012-09-01 00:00 Write a doc for TheBot\n'
                'cd) 2012-10-05 00:00 Celebrate my birthday\n'
                '9c) 2012-12-18 00:00 Celebrate daughter\'s birthday',
            ],
            adapter._lines
        )


def test_todo_plugin_for_different_users():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:
        adapter = bot.get_adapter('test')

        adapter.write('remind at 2012-10-05 to Celebrate my birthday', user='blah')
        adapter.write('remind at 2012-12-18 to Celebrate daughter\'s birthday', user='minor')

        adapter._lines[:] = []
        adapter.write('my tasks', user='minor')

        eq_(
            [
                '9c) 2012-12-18 00:00 Celebrate daughter\'s birthday',
            ],
            adapter._lines
        )


def test_todo_remind():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:
        adapter = bot.get_adapter('test')
        plugin = bot.get_plugin('todo')


        adapter.write('set my timezone to Asia/Shanghai')
        # these are the local times
        adapter.write('remind at 2012-09-05 10:00 to do task1')
        adapter.write('remind at 2012-09-05 10:30 to do task2')

        with mock.patch.object(times, 'now') as now:
            # this is a server time in UTC
            # it is 10:01 at Shanghai (+8 hours)
            now.return_value = datetime.datetime(2012, 9, 5, 2, 1)

            adapter._lines[:] = []
            plugin._remind_users_about_their_tasks()

            eq_(['TODO: do task1 (03f9)'], adapter._lines)

            # but it does not reminds twice
            now.return_value = datetime.datetime(2012, 9, 5, 2, 12)

            adapter._lines[:] = []
            plugin._remind_users_about_their_tasks()
            eq_([], adapter._lines)


def test_todo_done():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:

        adapter = bot.get_adapter('test')


        adapter.write('set my timezone to Asia/Shanghai')
        adapter.write('remind at 2012-09-05 10:00 to do task1')
        adapter.write('remind at 2012-09-05 10:30 to do task2')
        adapter.write('03 done')

        adapter._lines[:] = []
        adapter.write('my tasks')

        eq_(
            [
                '26) 2012-09-05 10:30 do task2',
            ],
           adapter._lines
        )


def test_todo_remind_at_uses_timezones():
    with closing(Bot(adapters=[TestAdapter], plugins=[todo.Plugin])) as bot:
        adapter = bot.get_adapter('test')
        plugin = bot.get_plugin('todo')


        adapter.write('set my timezone to Asia/Shanghai')
        adapter.write('remind at 2012-09-05 00:00 to do task1')
        tasks = plugin._get_tasks('some user')
        eq_(datetime.datetime(2012, 9, 4, 16, 0), tasks[0][0])


def test_yaml_config():
    cfg = Config()
    cfg.read_from_string("""
adapters: [xmpp, irc]
irc:
    channels: [thebot, test]
xmpp:
    jid: thebot@ya.ru
""")

    eq_(['xmpp', 'irc'], cfg.adapters)
    eq_('thebot@ya.ru', cfg.xmpp_jid)




# coding: utf-8

import thebot
import bisect
import times
import hashlib
import six

from dateutil.parser import parse
from collections import defaultdict


def _gen_hashes(tasks):
    """Generates sha1 hashes for tasks and minumum subhash's
    length for which all generated hashes are unique.
    """
    if not tasks:
        return [], 0

    shas = [
        hashlib.sha1(about.encode('utf-8')).hexdigest()
        for dt, about, r in tasks
    ]
    for min_len in range(2, len(shas[0]) + 1):
        trimmed = set(h[:min_len] for h in shas)
        if len(trimmed) == len(shas):
            break
    return shas, min_len


class Plugin(thebot.ThreadedPlugin):
    name = 'todo'

    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)

        # interval between bot's attemts to find expired tasks
        self.interval = 60

        if not self.bot.config.unittest:
            self.start_worker(interval=self.interval)

    def _get_tasks(self, user):
        return self.storage.get('tasks', defaultdict(list))[user]

    def _set_tasks(self, user, tasks):
        all_tasks = self.storage.get('tasks', defaultdict(list))
        all_tasks[user] = tasks
        self.storage['tasks'] = all_tasks

    @thebot.route(six.u('remind( me)? at (?P<datetime>.+) to (?P<about>.+)'))
    def remind(self, request, datetime, about):
        """Remind about a TODO at given time."""

        try:
            user = request.get_user()
            dt = parse(datetime)
            tz = self._get_user_timezone(user)
            dt = times.to_universal(dt, tz)

            tasks = self._get_tasks(user)
            bisect.insort(tasks, (dt, about, request))
            self._set_tasks(user, tasks)
        except Exception as e:
            request.respond(six.u('Unable to parse a date: ') + six.text_type(e))
            raise

        request.respond(six.u('ok'))

    @thebot.route(six.u('my tasks'))
    def my_tasks(self, request):
        """Show my tasks"""
        user = request.get_user()
        tasks = self._get_tasks(user)

        if tasks:
            tz = self._get_user_timezone(user)
            hashes, min_len = _gen_hashes(tasks)

            lines = []
            for h, (dt, about, request) in zip(hashes, tasks):
                dt = times.to_local(dt, tz)
                lines.append(six.u('{0}) {1:%Y-%m-%d %H:%M} {2}').format(h[:min_len], dt, about))

            request.respond(six.u('\n').join(lines))
        else:
            request.respond(six.u('You have no tasks'))

    @thebot.route('(?P<task_id>[0-9a-z]{2,40}) done')
    def done(self, request, task_id):
        # TODO add unittests for two cases:
        # * when task_id not found
        # * when there are more than one task which hash started from task_id
        tasks = self._get_tasks(request.get_user())

        if tasks:
            hashes, min_len = _gen_hashes(tasks)
            filtered_tasks = []
            for h, (dt, about, r) in zip(hashes, tasks):
                if not h.startswith(task_id):
                    filtered_tasks.append((dt, about, r))

            self._set_tasks(request.get_user(), filtered_tasks)
        request.respond('done')

    def _remind_users_about_their_tasks(self):
        now = times.now()
        for todos in self.storage['tasks'].values():
            idx = bisect.bisect_left(todos, (now, None, None))
            to_remind = todos[:idx]

            for td, about, request in to_remind:
                delta = (now - td)
                if delta.seconds <= self.interval:
                    # Remind only if reminder's datetime is between
                    # this and previous checks
                    request.respond(six.u('TODO: {0} ({1})').format(
                        about,
                        hashlib.sha1(about.encode('utf-8')).hexdigest()[:4]
                    ))

    def do_job(self):
        self._remind_users_about_their_tasks()

    @thebot.route(six.u('set my timezone to (?P<timezone>.+/.+)'))
    def set_timezone(self, request, timezone):
        """Set you timezone"""
        timezones = self.storage.get('timezones', {})
        timezones[request.get_user()] = timezone
        self.storage['timezones'] = timezones

        request.respond('done')

    def _get_user_timezone(self, user):
        timezones = self.storage.get('timezones', {})
        return timezones.get(user, 'UTC')


    @thebot.route(six.u('now'))
    def now(self, request):
        """Outputs server time and user time."""

        now = times.now()
        user = request.get_user()
        tz = self._get_user_timezone(user)
        local = times.to_local(now, tz)

        request.respond(six.u('Server time: {}\nLocal time:{}').format(now, local))



import sys

if sys.version_info.major == 3:
    force_unicode = lambda x: x
    force_str = lambda x: x
else:
    def force_unicode(value):
        if isinstance(value, str):
            value = value.decode('utf-8')
        return value

    def force_str(value):
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        return value



from __future__ import absolute_import

import re
import irc
import logging
import thebot
import threading


class IRCRequest(thebot.Request):
    def __init__(self, message, bot, nick, channel):
        super(IRCRequest, self).__init__(message)
        self.bot = bot
        self.nick = nick
        self.channel = channel

    def get_user(self):
        return (self.channel, self.nick)

    def respond(self, message):
        message = thebot.utils.force_str(message)
        adapter = self.bot.get_adapter('irc')
        irc_connection = adapter.irc_connection

        for line in message.split('\n'):
            irc_connection.respond(line, channel=self.channel, nick=self.nick)


class IRCConnection(irc.IRCConnection):
    def get_logger(self, logger_name, filename):
        """We override this method because don't want to have a separate log for irc messages.
        """
        return logging.getLogger(logger_name)


class Adapter(thebot.Adapter):
    @staticmethod
    def get_options(parser):
        group = parser.add_argument_group('IRC options')
        group.add_argument(
            '--irc-host', default='irc.freenode.net',
            help='Server to connect. Default: irc.freenode.net.'
        )
        group.add_argument(
            '--irc-port', default=6667,
            help='Port to connect. Default: 6667.'
        )
        group.add_argument(
            '--irc-channels', default='thebot',
            help='Comma-separated list of channels. Default: thebot.',
        )
        group.add_argument(
            '--irc-nick', default='thebot',
            help='IRC nick. Default: thebot.',
        )


    def start(self):
        thread = threading.Thread(target=self.run_bot)
        thread.daemon = True
        thread.start()

    def run_bot(self):
        """
        Convenience function to start a bot on the given network, optionally joining
        some channels
        """
        host = self.bot.config.irc_host
        port = self.bot.config.irc_port
        nick = self.bot.config.irc_nick
        channels = self.bot.config.irc_channels.split(',')

        def on_message(nick, message, channel):
            """A callback to be called by irckit when new message will arrive.

            It verifies if a bot's nick is mentioned, and pass data to TheBot.
            """
            message = thebot.utils.force_unicode(message)
            nick_re = re.compile(u'^%s[:,\s]\s*' % conn.nick)

            if nick_re.match(message) is not None:
                message = nick_re.sub(u'', message)
                request = IRCRequest(message, self.bot, nick, channel)
                return self.callback(request)

        conn = IRCConnection(host, port, nick)
        conn.register_callbacks((
            (re.compile(u'.*'), on_message),
        ))

        self.irc_connection = conn

        while 1:
            conn.connect()
            channels = channels or []
            for channel in channels:
                conn.join(channel)
            conn.enter_event_loop()





from __future__ import absolute_import

import irc
import logging
import thebot
import threading

from wsgiref.simple_server import make_server, WSGIRequestHandler


class HttpRequest(thebot.Request):
    def __init__(self, environ, start_response):
        super(HttpRequest, self).__init__(environ['PATH_INFO'])
        self.environ = environ
        self.start_response = start_response
        self.response_sent = False

    def respond(self, message):
        if self.response_sent:
            raise RuntimeError('Response to this HTTP request already sent.')

        status = '200 OK'
        headers = [
            ('Server', 'TheBot/' + thebot.__version__),
            ('Content-type', 'text/plain; charset=utf-8'),
        ]

        write = self.start_response(status, headers)
        write(message)
        self.response_sent = True


class IRCConnection(irc.IRCConnection):
    def get_logger(self, logger_name, filename):
        """We override this method because don't want to have a separate log for irc messages.
        """
        return logging.getLogger(logger_name)


class Adapter(thebot.Adapter):
    @staticmethod
    def get_options(parser):
        group = parser.add_argument_group('IRC options')
        group.add_argument(
            '--http-host', default='127.0.0.1',
            help='IP to bind to. Default: 127.0.0.1.'
        )
        group.add_argument(
            '--http-port', default=8888,
            help='TCP port to bind to. Default: 8888.'
        )


    def start(self):
        class QuietHandler(WSGIRequestHandler):
            def __init__(self, *args, **kwargs):
                self.logger = logging.getLogger('thebot.batteries.http')
                WSGIRequestHandler.__init__(self, *args, **kwargs)

            def log_message(self, format, *args):
                self.logger.info(
                    '%s - - [%s] %s',
                    self.address_string(),
                    self.log_date_time_string(),
                    format % args
                )

        server = make_server(
            self.bot.config.http_host,
            int(self.bot.config.http_port),
            self._wsgi_handler,
            handler_class=QuietHandler,
        )

        thread = threading.Thread(target=server.serve_forever)
        thread.daemon = True
        thread.start()

    def _wsgi_handler(self, environ, start_response):
        request = HttpRequest(environ, start_response)
        self.callback(request)
        if not request.response_sent:
            request.respond('')
        return []




from __future__ import absolute_import

import sleekxmpp
import thebot
import threading


class XMPPRequest(thebot.Request):
    def __init__(self, message, bot, _from):
        super(XMPPRequest, self).__init__(message)
        self.bot = bot
        self._from = _from

    def get_user(self):
        return self._from

    def respond(self, message):
        msg = sleekxmpp.stanza.message.Message()
        msg['to'] = self._from
        msg['type'] = 'chat'
        msg['body'] = message

        adapter = self.bot.get_adapter('xmpp')
        adapter.xmpp_bot.send(msg)


class Adapter(thebot.Adapter):
    @staticmethod
    def get_options(parser):
        group = parser.add_argument_group('XMPP options')
        group.add_argument(
            '--xmpp-jid', default='thebot@ya.ru',
            help='Jabber JID. Default: thebot@ya.ru.'
        )
        group.add_argument(
            '--xmpp-server', default='',
            help='Jabber server. Optional.'
        )
        group.add_argument(
            '--xmpp-password', default='',
            help='Password to connect to the server. Default: "".'
        )

    def start(self):
        thread = threading.Thread(target=self.run_bot)
        thread.daemon = True
        thread.start()

    def run_bot(self):
        """
        Convenience function to start a bot on the given network, optionally joining
        some channels
        """

        jid = self.bot.config.xmpp_jid
        password = self.bot.config.xmpp_password

        def on_message(msg):
            """A callback to be called by xmpppy's when new message will arrive.

            In it's turn, it will call TheBot's callback, to pass request into it.
            """

            if msg['type'] in ('chat', 'normal'):
                if msg['from'] != msg['to']:
                    request = XMPPRequest(msg['body'], self.bot, unicode(msg['from']))
                    self.callback(request)


        def on_start(event):
            self.xmpp_bot.get_roster()
            self.xmpp_bot.send_presence()


        self.xmpp_bot = sleekxmpp.ClientXMPP(jid, password)
        self.xmpp_bot._use_daemons = True
        self.xmpp_bot.add_event_handler('session_start', on_start)
        self.xmpp_bot.add_event_handler('message', on_message)

        self.xmpp_bot.connect()
        self.xmpp_bot.process(block=True)





# coding: utf-8

import requests
import anyjson
import random
import thebot


class Plugin(thebot.Plugin):
    @thebot.route('(image|img)( me)? (?P<query>.+)')
    def image(self, request, query):
        """Google random image on given topic."""
        url = self._find_image(query)
        if url is None:
            request.respond('No image was found for query "{0}"'.format(query))
        else:
            request.respond(url)

    @thebot.route('(?:mo?u)?sta(?:s|c)he?(?: for)? (?P<query>.+)')
    @thebot.route('усы для (?P<query>.+)')
    def mustache(self, request, query):
        """Apply moustache on image URL or random image on given topic."""
        type = int(random.randint(0, 2))

        if query.startswith('http'):
            url = query
        else:
            url = self._find_image(query)
            if url is None:
                request.respond('No image was found for query "{0}"'.format(query))
                return

        request.respond(
            'http://mustachify.me/{type}?src={url}'.format(
                type=type,
                url=url,
            )
        )

    def _find_image(self, query):
        response = requests.get(
            'http://ajax.googleapis.com/ajax/services/search/images',
            params=dict(
                v="1.0",
                rsz='8',
                q=query,
                safe='active',
            )
        )
        if response.status_code == 200:
            content = response.content
            data = anyjson.deserialize(content)
            images = data['responseData']['results']

            if len(images) > 0:
                image = random.choice(images[:5])
                return image['unescapedUrl']




from setuptools import setup, find_packages

setup(
    name='thebot',
    version='0.1.3',
    description=(
    ),
    keywords='chat irc xmpp basecamp jira fun',
    license = 'New BSD License',
    author="Alexander Artemenko",
    author_email='svetlyak.40wt@gmail.com',
    url='http://github.com/svetlyak40wt/thebot/',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    packages=find_packages(),
    scripts=['scripts/thebot'],
    install_requires = open('requirements/base.txt').readlines()
)



# coding: utf-8

import requests
import anyjson
import thebot


class Plugin(thebot.Plugin):
    @thebot.route('(calc|calculate|convert|math)( me)? (?P<expression>.+)')
    def math(self, request, expression):
        """Use Google's calculator to do some math."""
        response = requests.get(
            'http://www.google.com/ig/calculator',
            params=dict(
                q=expression,
            ),
        )
        if response.status_code == 200:
            content = response.content.decode(response.encoding)
            content = content.replace('{', '{"').replace(':', '":').replace(',', ',"')

            try:
                data = anyjson.deserialize(content)
            except Exception:
                request.respond('Can\'t parse: {}'.format(content))
            else:
                if data['rhs']:
                    request.respond(data['rhs'])
                else:
                    request.respond('Could not compute.')




import sys
import threading
import thebot


class ConsoleRequest(thebot.Request):
    def respond(self, message):
        sys.stdout.write('{0}\n'.format(thebot.utils.force_str(message)))
        sys.stdout.flush()


class Adapter(thebot.Adapter):
    def start(self):
        def loop():
            while True:
                sys.stdout.write('> ')
                sys.stdout.flush()
                line = sys.stdin.readline()
                if len(line) == 0:
                    # seems, Ctrl-D was pressed
                    self.callback(thebot.EXIT)
                    return

                line = line.strip()

                self.callback(ConsoleRequest(line))

        thread = threading.Thread(target=loop)
        thread.daemon = True
        thread.start()



